/* A simple tool to update the I2C EEPROM attached to the WR PTP core from a Linux terminal. */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "vfchdlib.h"
#include "wrc_syscon_regs.h"

#define I2C_EEPROM_SIZE ((size_t)(8 * 1024))
#define I2C_EEPROM_ADDR 0x51

void set_sda(struct vfchd_dev *dev, int val)
{
	if (val)
		vfchd_wrc_writel(dev, SYSC_GPSR_FMC_SDA,
				 dev->wrc_syscon_base + SYSC_REG_GPSR);
	else
		vfchd_wrc_writel(dev ,SYSC_GPCR_FMC_SDA,
				 dev->wrc_syscon_base + SYSC_REG_GPCR);
	usleep(1.25); /* 400kHz -> 2.5us/loop */
}

void set_scl(struct vfchd_dev *dev, int val)
{
	if (val)
		vfchd_wrc_writel(dev, SYSC_GPSR_FMC_SCL,
				 dev->wrc_syscon_base + SYSC_REG_GPSR);
	else
		vfchd_wrc_writel(dev, SYSC_GPCR_FMC_SCL,
				 dev->wrc_syscon_base + SYSC_REG_GPCR);
	usleep(1.25); /* 400kHz -> 2.5us/loop */
}

int get_sda(struct vfchd_dev *dev)
{
	return vfchd_wrc_readl(dev, dev->wrc_syscon_base + SYSC_REG_GPSR) &
		SYSC_GPSR_FMC_SDA ? 1 : 0;
}

static void mi2c_start(struct vfchd_dev *dev)
{
	set_sda(dev, 0);
	set_scl(dev, 0);
}

static void mi2c_stop(struct vfchd_dev *dev)
{
	set_sda(dev, 0);
	set_scl(dev, 1);
	set_sda(dev, 1);
}

static int mi2c_put_byte(struct vfchd_dev *dev, int data)
{
	int i;
	int ack;

	for (i = 0; i < 8; i++, data<<=1) {
		set_sda(dev, data & 0x80);
		set_scl(dev, 1);
		set_scl(dev, 0);
	}

	set_sda(dev, 1);
	set_scl(dev, 1);

	ack = get_sda(dev);

	set_scl(dev, 0);
	set_sda(dev, 0);

	return ack ? -EIO : 0; /* ack low == success */
}

static int mi2c_get_byte(struct vfchd_dev *dev, unsigned char *data, int ack)
{
	int i;
	int indata = 0;

	/* assert: scl is low */
	set_scl(dev, 0);
	set_sda(dev, 1);
	for (i = 0; i < 8; i++) {
		set_scl(dev, 1);
		indata <<= 1;
		if (get_sda(dev))
			indata |= 0x01;
		set_scl(dev, 0);
	}

	set_sda(dev, (ack ? 0 : 1));
	set_scl(dev, 1);
	set_scl(dev, 0);
	set_sda(dev, 0);

	*data= indata;
	return 0;
}

void mi2c_init(struct vfchd_dev *dev)
{
	set_scl(dev, 1);
	set_sda(dev, 1);
}

int vfchd_eeprom_read(struct vfchd_dev *dev, uint16_t offset, void *buf, size_t size)
{
	int ret = size;
	uint8_t *buf8 = buf;
	unsigned char c;

	if (offset > I2C_EEPROM_SIZE)
		return -EINVAL;
	if (offset + size > I2C_EEPROM_SIZE)
		return -EINVAL;

	/* Read it all in a single loop: hardware allows it */
	mi2c_start(dev);
	if(mi2c_put_byte(dev, I2C_EEPROM_ADDR << 1) < 0) {
		mi2c_stop(dev);
		return -EIO;
	}
	mi2c_put_byte(dev, (offset >> 8) & 0xff);
	mi2c_put_byte(dev, offset & 0xff);
	mi2c_stop(dev);
	mi2c_start(dev);
	mi2c_put_byte(dev, (I2C_EEPROM_ADDR << 1) | 1);
	while (size--) {
		mi2c_get_byte(dev, &c, size != 0);
		*buf8++ = c;
	}
	mi2c_stop(dev);
	return ret;
}

int vfchd_eeprom_write(struct vfchd_dev *dev, uint16_t offset, const void *buf, size_t size)
{
	int i, busy;
	const uint8_t *buf8 = buf;

	if (offset > I2C_EEPROM_SIZE)
		return -EINVAL;
	if (offset + size > I2C_EEPROM_SIZE)
		return -EINVAL;

	for(i = 0; i < size; i++) {

		mi2c_start(dev);

		if(mi2c_put_byte(dev, I2C_EEPROM_ADDR << 1) < 0) {
			mi2c_stop(dev);
			return -1;
		}
		mi2c_put_byte(dev, (offset >> 8) & 0xff);
		mi2c_put_byte(dev, offset & 0xff);
		mi2c_put_byte(dev, buf8[i]);
		mi2c_stop(dev);

		do { /* wait until the chip becomes ready */
			mi2c_start(dev);
			busy = mi2c_put_byte(dev, I2C_EEPROM_ADDR << 1);
			mi2c_stop(dev);
		} while(busy);
	}
	return size;
}

void vfchd_eeprom_dump (struct vfchd_dev *dev)
{

	uint8_t buf8[16];

	int i;

	for (i = 0; i < I2C_EEPROM_SIZE; i += 16) {

		if (vfchd_eeprom_read(dev, i, (void *)buf8, 16)) {

			printf ("%8.x: "
				"%.2x%.2x %.2x%.2x %.2x%.2x %.2x%.2x "
				"%.2x%.2x %.2x%.2x %.2x%.2x %.2x%.2x\n", i,
				buf8[0], buf8[1], buf8[2], buf8[3],
				buf8[4], buf8[5], buf8[6], buf8[7],
				buf8[8], buf8[9], buf8[10], buf8[11],
				buf8[12], buf8[13], buf8[14], buf8[15]);
		} else {
			fprintf (stderr, "(EE): error reading from address 0x%.4x\n", i);
		}

	}
}

int vfchd_eeprom_loader (struct vfchd_dev *dev, const char *fname)
{

	int retval = 0;

	FILE *f = fopen(fname, "rb");
	if ( f == NULL )
	{
		fprintf( stderr, "(EE): Could not access file %s\n", fname );
		return -1;
	}

	fseek(f, 0, SEEK_END);

	long fsize = ftell(f);

	fseek(f, 0, SEEK_SET);

	uint8_t datum, readback;

	int i = 0;

	vfchd_wrc_disable(dev);

	for (i=0; i < fsize; i++) {

		fread((void *)&datum, 1, 1, f);

		vfchd_eeprom_write (dev, i, &datum, 1);

		vfchd_eeprom_read (dev, i, &readback, 1);

		if ( readback != datum) {
			fprintf(stderr,
				"(EE): error writing value 0x%.2x to address "
				"0x%.4x. Read-back value was 0x%.2x\n",
				datum, i, readback);
			retval = -1;
			break;
		}

	}
	fclose(f);

	vfchd_wrc_enable(dev);

	return retval;
}

int main(int argc, char **argv)
{
	int retval = 0;

	struct vfchd_dev *dev = vfchd_open(argc, argv);

	if (!dev)
		return 1;

	if (optind >= argc) {
		printf("(II): no binary name provided, dumping current WRC EEPROM contents instead...\n");
		vfchd_eeprom_dump(dev);

	} else if (vfchd_eeprom_loader(dev, argv[optind]) < 0)	{
		fprintf(stderr, "(EE): WRC loader failure.\n");
		retval = 1;
	}

	vfchd_close(dev);

	return retval;
}
