/* A simple console for accessing the VFC-HD virtual UART (i.e. for communicating with the WR Core shell
   from a Linux terminal. */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>

#include "vfchdlib.h"
#include "wb_uart.h"

int vfchd_vuart_rx(struct vfchd_dev *dev, char *d)
{
	int rdr = vfchd_wrc_readl(dev, dev->wrc_vuart_base + UART_REG_HOST_RDR);

	if(rdr & UART_HOST_RDR_RDY)
	{
		*d = UART_HOST_RDR_DATA_R(rdr);
		return 1;
	}    else
		return 0;
}

int vfchd_vuart_tx(struct vfchd_dev *dev, char d)
{
	while( vfchd_wrc_readl(dev, dev->wrc_vuart_base + UART_REG_SR) & UART_SR_RX_RDY );
	vfchd_wrc_writel(dev, UART_HOST_TDR_DATA_W(d), dev->wrc_vuart_base + UART_REG_HOST_TDR);

	return 0;
}

static int transfer_byte(struct vfchd_dev *dev, int from, int is_control) {
	char c;
	int ret;
	do {
		ret = read(from, &c, 1);
	} while (ret < 0 && errno == EINTR);
	if(ret == 1) {
		if(is_control) {
			if(c == '\x01') { // C-a
				return -1;
			}
		}
		vfchd_vuart_tx(dev, c);
	} else {
		fprintf(stderr, "\n(EE): nothing to read. probably port disconnected.\n");
		return -2;
	}
	return 0;
}


void term_main(struct vfchd_dev *dev, int keep_term)
{
	struct termios oldkey, newkey;
	int need_exit = 0;

	fprintf(stderr, "[press C-a to exit]\n");

	if(!keep_term) {
		tcgetattr(STDIN_FILENO,&oldkey);
		newkey.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
		newkey.c_iflag = IGNPAR;
		newkey.c_oflag = 0;
		newkey.c_lflag = 0;
		newkey.c_cc[VMIN]=1;
		newkey.c_cc[VTIME]=0;
		tcflush(STDIN_FILENO, TCIFLUSH);
		tcsetattr(STDIN_FILENO,TCSANOW,&newkey);
	}
	while(!need_exit) {
		fd_set fds;
		int ret;
		char rx;
		struct timeval tv = {0, 10000};

		FD_ZERO(&fds);
		FD_SET(STDIN_FILENO, &fds);

		ret = select(STDIN_FILENO+1, &fds, NULL, NULL, &tv);
		if(ret == -1) {
			perror("select");
		} else if (ret > 0) {
			if(FD_ISSET(STDIN_FILENO, &fds)) {
				need_exit = transfer_byte(dev, STDIN_FILENO, 1);
			}
		}

		while((vfchd_vuart_rx(dev, &rx)) == 1)
			fprintf(stderr,"%c", rx);

	}

	if(!keep_term)
		tcsetattr(STDIN_FILENO,TCSANOW,&oldkey);
}

int main(int argc, char **argv)
{
	struct vfchd_dev *dev = vfchd_open(argc, argv);

	if (!dev)
		return 1;

	term_main(dev, 0);

	vfchd_close(dev);

	return 0;
}
