#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "vfchdlib.h"
#include "wrc_syscon_regs.h"

uint32_t _vfchd_readl(struct vfchd_dev *dev, uint32_t offset)
{
	volatile uint32_t *addr = dev->map.user_va + offset;

	uint32_t datum = ntohl(addr[0]);

	return datum;
}

void _vfchd_writel(struct vfchd_dev *dev, uint32_t datum, uint32_t offset)
{
	volatile uint32_t *addr = dev->map.user_va + offset;

	addr[0] = htonl(datum);
}

uint32_t vfchd_wrc_readl(struct vfchd_dev *dev, uint32_t offset)
{
	return _vfchd_readl(dev, dev->wrc_base_offset + offset);
}


void vfchd_wrc_writel(struct vfchd_dev *dev, uint32_t datum, uint32_t offset)
{
	_vfchd_writel(dev, datum, dev->wrc_base_offset + offset);
}

void vfchd_wrc_disable(struct vfchd_dev *dev)
{
	vfchd_wrc_writel(dev, 0x1deadbee, dev->wrc_syscon_base);
}

void vfchd_wrc_enable(struct vfchd_dev *dev)
{
	vfchd_wrc_writel(dev, 0x0deadbee, dev->wrc_syscon_base);
}

void _vfchd_print_usage(const char *progname)
{
	fprintf(stderr,
		"\nUsage: \"%s "
		"[-a VME_AM] "
		"[-d DATA_WIDTH] "
		"[-s MAP_SIZE] "
		"[-w WRC_BASE_OFFSET] "
		"[-m WRC_MEM_BASE] "
		"[-y WRC_SYSCON_BASE] "
		"[-u WRC_VUART_BASE] "
		"-l SLOT [...]\n", progname);
	fprintf(stderr,
		"By default, a=0x%02x, d=%d, "
		"s=0x%x, w=0x%x, "
		"m=0x%x, y=0x%x, "
		"and u=0x%x\n\n",
		VFCHD_AM_DEFAULT,
		VFCHD_DW_DEFAULT,
		VFCHD_SIZEL_DEFAULT,
		VFCHD_WRC_OFFSET_DEFAULT,
		VFCHD_WRC_MEMORY_DEFAULT,
		VFCHD_WRC_SYSCON_DEFAULT,
		VFCHD_WRC_VUART_DEFAULT);
}

int _vfchd_vme64x_rellocate(struct vfchd_dev *dev)
{
	/* temporary map, will be discarded at the end of the
	   rellocation function */
	dev->map.am         = VME_CR_CSR;
	dev->map.data_width = VME_D32;
	dev->map.sizel      = 0x80000;
	dev->map.vme_addrl  = dev->slot << 19;

	if (vme_map(&(dev->map), 1) == NULL) {
		fprintf(stderr, "(EE): could not map at 0x%08x\n",
			dev->map.vme_addrl);
		return -1;
	}

	/* reset the core */
	_vfchd_writel(dev, RESET_CORE, BIT_SET_REG);
	usleep(10000);

	/* disable the core */
	_vfchd_writel(dev, ENABLE_CORE, BIT_CLR_REG);

	/* default to 32bit WB interface */
	_vfchd_writel(dev, WB32, WB_32_64);

	/* set interrupt vector and level */
	_vfchd_writel(dev, 0, INTVECTOR);
	_vfchd_writel(dev, 0, INT_LEVEL);

	/* FUNCTION0 for A32 (dev->slot << 24) */
	_vfchd_writel(dev,  dev->slot        & 0xff, FUN0ADER);
	_vfchd_writel(dev, (dev->slot <<  8) & 0xff, FUN0ADER + 4);
	_vfchd_writel(dev, (dev->slot << 16) & 0xff, FUN0ADER + 8);
	_vfchd_writel(dev, (VME_A32_USER_DATA_SCT & 0x3f) << 2,
		      FUN0ADER + 12);

	/* FUNCTION1 for A24 (dev->slot << 19) */
	_vfchd_writel(dev, (dev->slot >>  5) & 0xff, FUN1ADER);
	_vfchd_writel(dev, (dev->slot <<  3) & 0xff, FUN1ADER + 4);
	_vfchd_writel(dev, (dev->slot << 11) & 0xff, FUN1ADER + 8);
	_vfchd_writel(dev, (VME_A24_USER_DATA_SCT & 0x3f) << 2,
		      FUN1ADER + 12);

	/* enable module, hence make FUN0/1 available */
	_vfchd_writel(dev, ENABLE_CORE, BIT_SET_REG);

	vme_unmap(&(dev->map), 1);

	return 0;
}

struct vfchd_dev *vfchd_open(int argc, char **argv)
{
	struct vfchd_dev *dev;

	int c;

	// Required arguments
	uint8_t slot = 0;

	// Default values for optional arguments
	uint32_t     wrc_base_offset = VFCHD_WRC_OFFSET_DEFAULT;
	uint32_t        wrc_mem_base = VFCHD_WRC_MEMORY_DEFAULT;
	uint32_t     wrc_syscon_base = VFCHD_WRC_SYSCON_DEFAULT;
	uint32_t      wrc_vuart_base = VFCHD_WRC_VUART_DEFAULT;
	enum vme_address_modifier am = VFCHD_AM_DEFAULT;
	enum vme_data_width       dw = VFCHD_DW_DEFAULT;
	unsigned int           sizel = VFCHD_SIZEL_DEFAULT;

	// Process cmd line arguments
	while ((c = getopt (argc, argv, "a:d:s:w:m:y:u:l:")) != -1)
		switch (c)
		{
		case 'a':
			am = strtol(optarg, NULL, 0);
			break;
		case 'd':
			dw = strtol(optarg, NULL, 0);
			break;
		case 's':
			sizel = strtol(optarg, NULL, 0);
			break;
		case 'w':
			wrc_base_offset = strtol(optarg, NULL, 0);
			break;
		case 'm':
			wrc_mem_base = strtol(optarg, NULL, 0);
			break;
		case 'y':
			wrc_syscon_base = strtol(optarg, NULL, 0);
			break;
		case 'u':
			wrc_vuart_base = strtol(optarg, NULL, 0);
			break;
		case 'l':
			slot  = strtol(optarg, NULL, 0);
			break;
		case '?':
		default:
			_vfchd_print_usage(argv[0]);
			return NULL;
		}

	if (!slot) {
		fprintf(stderr, "(EE): missing SLOT argument.\n");
		_vfchd_print_usage(argv[0]);
		return NULL;
	}

	// Allocate and init the device structure
	dev = malloc(sizeof(struct vfchd_dev));

	if (!dev)
		return NULL;

	dev->wrc_base_offset = wrc_base_offset;
	dev->wrc_mem_base    = wrc_mem_base;
	dev->wrc_syscon_base = wrc_syscon_base;
	dev->wrc_vuart_base  = wrc_vuart_base;

	dev->slot = slot;

	// Perform VME64x rellocation if necessary
	if (_vfchd_vme64x_rellocate(dev) < 0) {
		fprintf(stderr,"(EE): VME64x rellocation failed.\n");
		free(dev);
		return NULL;
	}

	dev->map.am         = am;
	dev->map.data_width = dw;
	dev->map.sizel      = sizel;

	switch(am)
	{
	case VME_A32_USER_DATA_SCT:
		dev->map.vme_addrl = dev->slot << 24;
		break;

	case VME_A24_USER_DATA_SCT:
		dev->map.vme_addrl = dev->slot << 19;
		break;

	default:
		dev->map.vme_addrl = 0;
		break;
	}

	if (!dev->map.vme_addrl) {
		fprintf(stderr,
			"(EE): unsupported AM 0x%02x, "
			"only 0x09 and 0x39 are allowed\n",
			dev->map.am);
		free(dev);
		return NULL;
	}

	if (vme_map(&(dev->map), 1) == NULL) {
		fprintf(stderr, "(EE): could not map at 0x%08x\n",
			dev->map.vme_addrl);
		free(dev);
		return NULL;
	}

	return (struct vfchd_dev *)dev;

}

void vfchd_close(struct vfchd_dev *dev) {

	if (!dev)
		return;

	vme_unmap(&(dev->map), 1);

	free(dev);
}
